package fr.username404.snowygui.gui.elements

import com.mojang.blaze3d.systems.RenderSystem
import com.mojang.blaze3d.vertex.DefaultVertexFormat
import com.mojang.blaze3d.vertex.PoseStack
import fr.username404.snowygui.config.Configuration
import fr.username404.snowygui.gui.ColoredElement
import fr.username404.snowygui.gui.Renderable.Rendering.buffer
import fr.username404.snowygui.gui.Renderable.Rendering.defaultRectFunc
import fr.username404.snowygui.gui.Renderable.Rendering.tessellator
import fr.username404.snowygui.gui.feature.ButtonImpl
import fr.username404.snowygui.gui.feature.Category
import fr.username404.snowygui.gui.feature.Colors
import io.github.config4k.extract
import kotlinx.coroutines.launch
import kotlinx.coroutines.runBlocking
import net.minecraft.client.Minecraft
import net.minecraft.network.chat.TranslatableComponent
import org.jetbrains.annotations.ApiStatus
import org.lwjgl.opengl.GL20

@ApiStatus.Internal
class ClickBox(
    x: Double, y: Double,
    override var color: Int = Colors.BLUE.hexValue,
    val name: TranslatableComponent? = null
): ColoredElement(x, y, 80, 10, color, 0.5F) {
    constructor(x: Double, y: Double, color: Colors, name: TranslatableComponent? = null): this(x, y, color.hexValue, name)

    fun isCategory(c: Category): Boolean = (name?.key == c.translationKey)
    val buttons: MutableSet<ButtonImpl> = if (sortAlphabetically) sortedSetOf(compareBy(String.CASE_INSENSITIVE_ORDER) { it.title }) else mutableSetOf()
    override fun display(stack: PoseStack?) {
        hidden = buttons.isEmpty() || hidden
        super.display(stack)
    }

    @JvmField
    val buttonsProgressBar: ColoredElement = object: ColoredElement(
        (x + width), y + height + 3,
        width = 3, height = clickboxHeightOffset - 6,
        opacity = 0.75F
    ) {
        override val color: Int get() = this@ClickBox.color
        override fun render(poseStack: PoseStack?) {
            RenderSystem.disableTexture()
            RenderSystem.enableBlend()
            RenderSystem.defaultBlendFunc()
            defaultRectFunc()
            RenderSystem.disableBlend()
            RenderSystem.enableTexture()
        }
        init { height = 8 }
    }
    var barStage: Int = 1; private set
    fun scroll(d: Double, e: Double, supplied: Double) {
        if (withinBounds(d, e, offsetHeight = clickboxHeightOffset.toDouble())) {
            with(buttonsProgressBar) {
                if ((height > 8 || (supplied < 0)) && (((height < (buttons.size * 8)) && (height < originalHeight)) || (supplied > 0))) {
                    height -= supplied.toInt()
                    if (buttons.isNotEmpty()) (height / 8).let {
                        if (it > 0) barStage = it
                    }
                }
            }
        }
    }

    companion object {
        private val savedColors = "box_colors".let {
            with(Configuration()) {
                if (!hasPath(it)) null else extract<Map<String, Int>>(it)
            }
        }
        var sortAlphabetically: Boolean by Configuration
        const val buttonsMax: Short = 16 // TODO Remove the buttons limit
        const val clickboxHeightOffset: Int = 80
        private const val inclination: Double = 2.5
    }
    override fun render(poseStack: PoseStack?) {
        runBlocking {
            RenderSystem.disableTexture()
            RenderSystem.enableBlend()
            RenderSystem.defaultBlendFunc()
            with(buffer) {
                begin(GL20.GL_POLYGON, DefaultVertexFormat.POSITION_COLOR)
                // Render the header:
                vertex(x, y + height, 0.0).colorEnd()
                vertex(x + width + inclination, y + height, 0.0).colorEnd()
                vertex(x + width, y, 0.0).colorEnd()
                vertex(x + inclination, y, 0.0).colorEnd()

                // Render the box:
                val currentHeight = y + (height + clickboxHeightOffset)
                vertex(x, currentHeight, 0.0).colorEnd()
                vertex(x + width + inclination, currentHeight, 0.0).colorEnd()
                vertex(x + width + inclination, y + height, 0.0).colorEnd()
                tessellator.end()

                begin(GL20.GL_LINES, DefaultVertexFormat.POSITION_COLOR)
                vertex(x + inclination, y + height, 0.0).colorEnd(Colors.WHITE_LINES.hexValue)
                vertex(x + width, y + height, 0.0).colorEnd(Colors.WHITE_LINES.hexValue)
                tessellator.end()
            }
            RenderSystem.enableTexture()
            RenderSystem.disableBlend()

            val renderButtons = if (buttons.isNotEmpty()) launch {
                buttonsProgressBar.apply {
                    x = this@ClickBox.x + this@ClickBox.width - 3
                    y = this@ClickBox.y + this@ClickBox.height + 3
                }.display(poseStack)
                buttons.forEachIndexed { num, button ->
                    val fullHeight = (y + height.toDouble())..(y + height + clickboxHeightOffset)
                    button.also {
                        it.x = x + 3
                        it.y = y + 3 + height + (((num + 1) - barStage) * 9)
                        it.hidden = if ((num + 1) <= 8) ((it.y) !in fullHeight) else ((it.y + it.height) !in fullHeight)
                    }.display(poseStack)
                }
            } else null
            if ((name != null) && (poseStack != null)) {
                Minecraft.getInstance().font.draw(poseStack, name.string, x.toFloat() + 5, y.toFloat() + 1.5F, Colors.TRANSPARENT.hexValue)
                renderButtons?.join()
            }
        }
    }
    init {
        savedColors?.get(name?.key)?.let {
            this.color = it
        }
    }
}
